@@ -6,17 +6,20 @@
     See the file COPYING.LIB.
 */
 
-#include "fuse.h"
+#include "fuse_i.h"
 #include "fuse_opt.h"
 
+#include <sys/stat.h>
+#include <sys/wait.h>
+#include <sys/sysctl.h>
 #include <stdio.h>
 #include <stdlib.h>
 #include <unistd.h>
 #include <stddef.h>
 #include <fcntl.h>
 #include <errno.h>
-#include <sys/wait.h>
 #include <string.h>
+#include <paths.h>
 
 #define FUSERMOUNT_PROG         "mount_fusefs"
 #define FUSE_DEV_TRUNK          "/dev/fuse"
@@ -69,6 +72,13 @@ static const struct fuse_opt fuse_mount_opts[] = {
     FUSE_OPT_KEY("ro",                  KEY_KERN),
     FUSE_OPT_KEY("rw",                  KEY_KERN),
     FUSE_OPT_KEY("auto",                KEY_KERN),
+    /* options supported under both Linux and FBSD */
+    FUSE_OPT_KEY("allow_other",         KEY_KERN),
+    FUSE_OPT_KEY("default_permissions", KEY_KERN),
+    /* FBSD FUSE specific mount options */
+    FUSE_OPT_KEY("private",             KEY_KERN),
+    FUSE_OPT_KEY("neglect_shares",      KEY_KERN),
+    FUSE_OPT_KEY("push_symlinks_in",    KEY_KERN),
     /* stock FBSD mountopt parsing routine lets anything be negated... */
     FUSE_OPT_KEY("nodev",               KEY_KERN),
     FUSE_OPT_KEY("noasync",             KEY_KERN),
@@ -93,13 +103,11 @@ static const struct fuse_opt fuse_mount_opts[] = {
     FUSE_OPT_KEY("noro",                KEY_KERN),
     FUSE_OPT_KEY("norw",                KEY_KERN),
     FUSE_OPT_KEY("noauto",              KEY_KERN),
-    /* options supported under both Linux and FBSD */
-    FUSE_OPT_KEY("allow_other",         KEY_KERN),
-    FUSE_OPT_KEY("default_permissions", KEY_KERN),
-    /* FBSD FUSE specific mount options */
-    FUSE_OPT_KEY("private",             KEY_KERN),
-    FUSE_OPT_KEY("neglect_shares",      KEY_KERN),
-    FUSE_OPT_KEY("push_symlinks_in",    KEY_KERN),
+    FUSE_OPT_KEY("noallow_other",       KEY_KERN),
+    FUSE_OPT_KEY("nodefault_permissions", KEY_KERN),
+    FUSE_OPT_KEY("noprivate",           KEY_KERN),
+    FUSE_OPT_KEY("noneglect_shares",    KEY_KERN),
+    FUSE_OPT_KEY("nopush_symlinks_in",  KEY_KERN),
     /* Linux specific mount options, but let just the mount util handle them */
     FUSE_OPT_KEY("fsname=",             KEY_KERN),
     FUSE_OPT_KEY("nonempty",            KEY_KERN),
@@ -154,7 +162,7 @@ static int fuse_mount_opt_proc(void *data, const char *arg, int key,
     return 1;
 }
 
-void fuse_unmount(const char *mountpoint)
+void fuse_unmount_compat22(const char *mountpoint)
 {
     char dev[128];
     char *ssc, *umount_cmd;
@@ -171,6 +179,14 @@ void fuse_unmount(const char *mountpoint)
 
     (void) mountpoint;
 
+    /*
+     * If we don't know the fd, we have to resort to the scripted solution --
+     * iterating over the fd-s is unpractical, as we don't know how many of
+     * open files we have. (This could be looked up in procfs -- however,
+     * that's optional on FBSD; or read out from the kmem --  however, that's
+     * bound to privileges (in fact, that's what happens when we call the
+     * setgid kmem fstat(1) utility).
+     */
     asprintf(&ssc, seekscript, getpid());
 
     errno = 0;
@@ -187,6 +203,43 @@ void fuse_unmount(const char *mountpoint)
     system(umount_cmd);
 }
 
+void fuse_kern_unmount(const char *mountpoint, int fd)
+{
+    char *ep, *umount_cmd, dev[128];
+    struct stat sbuf;
+
+    (void)mountpoint;
+
+    if (fstat(fd, &sbuf) == -1)
+        return;
+
+    devname_r(sbuf.st_rdev, S_IFCHR, dev, 128);
+
+    if (strncmp(dev, "fuse", 4))
+        return;
+
+    strtol(dev + 4, &ep, 10);
+    if (*ep != '\0')
+        return;
+
+    asprintf(&umount_cmd, "/sbin/umount " _PATH_DEV "%s", dev);
+    system(umount_cmd);
+}
+
+/* Check if kernel is doing init in background */
+static int init_backgrounded(void)
+{
+    int ibg, len;
+
+    len = sizeof(ibg);
+
+    if (sysctlbyname("vfs.fuse.init_backgrounded", &ibg, &len, NULL, 0))
+        return 0;
+
+    return ibg;
+}
+
+
 static int fuse_mount_core(const char *mountpoint, const char *opts)
 {
     const char *mountprog = FUSERMOUNT_PROG;
@@ -235,12 +288,19 @@ mount:
     }
 
     if (pid == 0) {
-        pid = fork();
+        if (! init_backgrounded()) {
+            /*
+             * If init is not backgrounded, we have to call the mount util
+             * backgrounded, to avoid deadlock.
+             */
 
-        if (pid == -1) {
-            perror("fuse: fork() failed");
-            close(fd);
-            exit(1);
+            pid = fork();
+    
+            if (pid == -1) {
+                perror("fuse: fork() failed");
+                close(fd);
+                exit(1);
+            }
         }
 
         if (pid == 0) {
@@ -272,7 +332,7 @@ out:
     return fd;
 }
 
-int fuse_mount(const char *mountpoint, struct fuse_args *args)
+int fuse_kern_mount(const char *mountpoint, struct fuse_args *args)
 {
     struct mount_opts mo;
     int res = -1;
@@ -299,3 +359,5 @@ int fuse_mount(const char *mountpoint, struct fuse_args *args)
     free(mo.kernel_opts);
     return res;
 }
+
+__asm__(".symver fuse_unmount_compat22,fuse_unmount@FUSE_2.2");
